	################################################################
	#
	# Minimalistinen Hei maailma -sovellus GNU Assemblerilla.
	#
	################################################################
	#
	# Tässä demonstroidaan käyttöjärjestelmäkutsua (syscall -käsky
        # AMD64 -arkkitehtuurissa). Näin se loppuviimein tapahtuu aina
	# jossakin matalimman tason alustakirjaston syövereissä. Tämä on
	# nyt täysin kiinni x86-64 -arkkitehtuurissa, Linuxissa ja GNU:n
	# assembler-syntaksissa.
	#
	# Tehdään suoritettavaksi ohjelmaksi seuraavilla GNU-työkalujen
	# komennoilla esim. suorakäyttökoneissamme:
	#
	#   as -o helloasm.o helloasm_kommentoitu.s
	#   ld -o helloasm helloasm.o
	#
	# Ensimmäinen käyttää assembler-kääntäjää komennolla 'as'
	# tuottaakseen objektitiedoston helloasm.o
	#
	# Toinen linkittää objektin helloasm.o suoritettavaksi tiedostoksi
	# nimeltä helloasm. Sitä voi sitten ajaa ja debugata luennon tapaan
	# seuraavilla komennoilla:
	#
	#   ./helloasm
	#   gdb helloasm
	#
	# Seuraavalla luennolla viimeistellään koko kuvion kulku eli:
	#
	#   lähdekoodi -> assembly -> objekti -> executable -> prosessi
	#
	################################################################
	

	# Julkaistaan myöhemmin määriteltävä symbolinen muistiosoite _start
	# "globaalisti", koska sen perusteella linkkeri tunnistaa,
	# mistä kohtaa alkaa aloituspisteeksi tarkoitettu ohjelmakoodi.
	
	.globl	_start
	
	# Kirjoitetaan koodisegmenttiin ("text") nyt sekä data että koodi.
	# Laajemmissa ohjelmissa olisi erikseen osiot ainakin koodille (text),
	# vakiodatalle (data) ja nollaksi alustettavalle työtilalle (bss).
	# Seuraava ohjerivi kertoo assemblerille, että nyt pitää alkaa
	# tuutata tavaraa koodialueelle:

	.text

.hei_mun_maailmani:                # (Symbolinen nimi muistiosoitteelle)
	.ascii	"Hel"              # Sisältöä merkkeinä. Tässä merkit
	.ascii	"lo world."        # kopioituvat ohjelmakoodiin peräkkäin
	.byte	0xa                # ... ja sekaan voi heittää lukuarvoja
	.ascii "jeijee!"           # ... kaikki menee vaan koodipötköksi
	.byte	0xa                # ... assembler on aika yksinkertaista!
	                           # Jos merkkijonon loppuun haluttaisiin
	                           # automaattisesti nollamerkki, voisi
				   # käyttää .ascii:n sijasta .string
				   # kuten luentoesimerkissä oli.
	
.hei_mun_maailmani_loppu:          # (Symbolinen nimi muistiosoitteelle)
		
_start:                            # (Symbolinen nimi muistiosoitteelle)

	# Yksi pikkujuttu, joka on tarpeen ainakin liukulukuja käsiteltäessä..
	# en oo ehtinyt selvittää, vaikuttaako, jos ei käytetä liukulukuja:
	and	$0xfffffffffffffff0,%rsp

	# (-> ei vaikuta toimintaan normaaleilla käskyillä, mutta on kyllä
	# erikseen ihan speksattu, että pinon täytyy olla jokaisen aliohjelma-
	# kutsun alkaessa 16 tavun rajalla, eli alimmat 4 bittiä nollia!)
	
	# Järjestelmäkutsuun mennään x86_64 -prosessoriarkkitehtuurissa
	# käskyllä syscall, joka aiheuttaa prosessorin "kevyen" keskeytyksen. 
	# Toimenpiteet on dokumentoitu esim. AMD64-arkkitehtuurimanuaalissa.
	# Eri prosessoreissa voi olla eri niminen ja hieman eri tavoin
	# toimiva käsky ns. ohjelmoidun keskeytyksen aikaansaamiseksi.
	#
	# Sen jälkeen prosessori suorittaa käyttöjärjestelmän koodia, joten
	# on täysin käyttöjärjestelmän toteutuksesta riippuvaa, mitä
	# ohjelmoidun keskeytyksen jälkeen alkaa tapahtua, ja mitä on pitänyt
	# tapahtua sovellusohjelman puolella, että toivottu käyttöjärjestelmän
	# palvelu saadaan tarkoitetulla tavoin käyttöön.
	#
	# Linuxin x86_64 -versio odottaa RAX-rekisterissä tietoa, mikä
	# nimenomainen järjestelmäkutsu sen pitäisi tehdä. Dokumentaation
	# perusteella se tulostaa kutsulla numero 1, joten Helloworldissä
	# sovelluksen on laitettava RAX-rekisteriin ykkönen.
	#
	# Kyseinen tulostuspalvelu odottaa juuri tietyissä rekistereissä
	# olevan tietyt parametrit kutsua varten:
        # 
	# RDI:ssä tulostuksen kohteena olevaa tiedostoa/tietovirtaa kuvaava
	# numero. Standardi ulostulovirta on avattu valmiiksi ennen ohjelman 
	# käynnistymistä numerokoodille 1, joten laitetaan se RDI:hin, kun
	# halutaan tulostaa standardiulostuloon. Olennaisesti samalla tavoin
	# tulostettaisiin myös kovalevylle, mutta ensin pitäisi avata tiedosto
	# uudelle numerokoodille - arvatenkin eri käyttöjärjestelmäkutsulla...
	# 
	# RSI:ssä täytyy olla tulostettavan datan alkuosoite muistissa. Meillä
	# on itse määritelty symbolinen nimi, jonka Assembler-kääntäjä osaa
	# muuntaa. Itse asiassa muistiosoitteen lopullinen numeroarvo saattaa
	# selvitä vasta siinä vaiheessa kun käyttöjärjestelmä lataa tätä
	# ohjelmaa suoritukseen ja yhdistelee ohjelman osioita toisiinsa
	# muistiavaruuden eri kohtiin. Onneksi voidaan käyttää symboleita.
	#
	# RDX:ssä täytyy olla tulostettavan datan määrä tavuina. Onneksi
	# voidaan laskea "pituus = loppuosoite - alkuosoite". Tästä tulee
	# konkreettinen lukuarvo jo käännösvaiheessa, koska data sijaitsee
	# peräkkäisissä muistiosoitteissa, eikä niiden keskinäinen
	# suhteellinen sijainti muutu ohjelman lataamisessa tai
	# linkittämisessä.
	#
	# Parametrit kun on laitettu kohdilleen, niin sitten tarvitsee vaan
	# pamauttaa prosessori keskeytyskäsittelyyn ja luottaa, että ollaan
	# suorituksessa Linux-käyttöjärjestelmän sellaisen version päällä,
	# jonka järjestelmäkutsurajapinnassa on samat numerot käytössä kuin
	# tätä ohjelmaa käännettäessä.
        #
	# Ja ymmärretään pari asiaa: (1) Konekieli, jota tietokone suorittaa,
	# on yksinkertaista (2) Assembler on yksinkertaista (3) Alimman tason
	# käyttöjärjestelmärajapinta on yksinkertainen (4) Kaikesta tästä
	# yksinkertaisuudesta johtuen ei ole juuri mitään järkeä käyttää 
	# laitetta ilman monitasoista abstraktiota, josta C-kielinen
	# Linux-rajapinta on seuraava ylempi taso, POSIX-standardin määräämä
	# yhteensopivien käyttöjärjestelmien C-kutsurajapinta sitä seuraava,
	# ja näiden päälle tietenkin löytyy erilaisia ohjelmointikieliä ja
	# kirjastoja, jotka edelleen abstrahoivat toimintaa
	# yleiskäyttöisemmäksi ja helpommaksi. (5) Johtopäätös:
	# Yksinkertainen ei tosiaankaan ole sama kuin helppokäyttöinen.

	movq 	$1, %rax
	movq    $1, %rdi
	movq	$.hei_mun_maailmani, %rsi
	movq	$.hei_mun_maailmani_loppu - .hei_mun_maailmani, %rdx
	syscall

	# Seuraava on helppo selittää edellisen sepustuksen jälkeen:
	# Linuxin x86_64 -versio lopettaa ohjelman kutsulla numero 60, joka
	# odottaa, että virhekoodiksi tarkoitettu luku on rekisterissä RDI.
	
	movq $60,%rax
	movq $0,%rdi
	syscall
